classnames = [
    "apple pie",
    "baby back ribs",
    "baklava",
    "beef carpaccio",
    "beef tartare",
    "beet salad",
    "beignets",
    "bibimbap",
    "bread pudding",
    "breakfast burrito",
    "bruschetta",
    "caesar salad",
    "cannoli",
    "caprese salad",
    "carrot cake",
    "ceviche",
    "cheese plate",
    "cheesecake",
    "chicken curry",
    "chicken quesadilla",
    "chicken wings",
    "chocolate cake",
    "chocolate mousse",
    "churros",
    "clam chowder",
    "club sandwich",
    "crab cakes",
    "creme brulee",
    "croque madame",
    "cup cakes",
    "deviled eggs",
    "donuts",
    "dumplings",
    "edamame",
    "eggs benedict",
    "escargots",
    "falafel",
    "filet mignon",
    "fish and chips",
    "foie gras",
    "french fries",
    "french onion soup",
    "french toast",
    "fried calamari",
    "fried rice",
    "frozen yogurt",
    "garlic bread",
    "gnocchi",
    "greek salad",
    "grilled cheese sandwich",
    "grilled salmon",
    "guacamole",
    "gyoza",
    "hamburger",
    "hot and sour soup",
    "hot dog",
    "huevos rancheros",
    "hummus",
    "ice cream",
    "lasagna",
    "lobster bisque",
    "lobster roll sandwich",
    "macaroni and cheese",
    "macarons",
    "miso soup",
    "mussels",
    "nachos",
    "omelette",
    "onion rings",
    "oysters",
    "pad thai",
    "paella",
    "pancakes",
    "panna cotta",
    "peking duck",
    "pho",
    "pizza",
    "pork chop",
    "poutine",
    "prime rib",
    "pulled pork sandwich",
    "ramen",
    "ravioli",
    "red velvet cake",
    "risotto",
    "samosa",
    "sashimi",
    "scallops",
    "seaweed salad",
    "shrimp and grits",
    "spaghetti bolognese",
    "spaghetti carbonara",
    "spring rolls",
    "steak",
    "strawberry shortcake",
    "sushi",
    "tacos",
    "takoyaki",
    "tiramisu",
    "tuna tartare",
    "waffles",
]

templates = [lambda c: f"a photo of {c}, a type of food."]
